import unittest
import pytest
from random import randint
from materials_commons.api import create_project, get_all_templates, Template


def fake_name(prefix):
    number = "%05d" % randint(0, 99999)
    return prefix + number


class TestMeasurementEPMASetup(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.project_name = fake_name("TestMeasurementEPMASetupProject-")
        description = "Test project generated by automated test"
        cls.project = create_project(cls.project_name, description)
        cls.project_id = cls.project.id
        name = "TestMeasurementEPMASetup"
        description = "Test experiment generated by automated test"
        cls.experiment = cls.project.create_experiment(name, description)
        cls.experiment_id = cls.experiment.id
        cls.process = cls.experiment.create_process_from_template(Template.create)
        cls.sample_name = "setup-sample-1"
        cls.samples = cls.process.create_samples(sample_names=[cls.sample_name])
        cls.sample = cls.samples[0]

    def test_is_setup_correctly(self):
        self.assertIsNotNone(self.project)
        self.assertIsNotNone(self.project.name)
        self.assertEqual(self.project_name, self.project.name)
        self.assertIsNotNone(self.project.id)
        self.assertEqual(self.project_id, self.project.id)
        self.assertIsNotNone(self.experiment)
        self.assertIsNotNone(self.experiment.id)
        self.assertEqual(self.experiment_id, self.experiment.id)
        self.assertIsNotNone(self.process)
        self.assertIsNotNone(self.process.id)
        self.assertIsNotNone(self.process.process_type)
        self.assertEqual(self.process.process_type, 'create')
        self.assertTrue(self.process.does_transform)
        sample = self.sample
        self.assertIsNotNone(sample)
        self.assertIsNotNone(sample.name)
        self.assertIsNotNone(sample.property_set_id)
        self.assertEqual(sample.name, self.sample_name)

    def test_fetch_sample_by_id(self):
        fetched_sample = self.project.fetch_sample_by_id(self.sample.id)
        self.assertIsNotNone(fetched_sample)
        self.assertIsNotNone(fetched_sample.processes)
        self.assertEqual(len(fetched_sample.processes), 1)
        process = fetched_sample.processes[0]
        self.assertIsNotNone(process.input_data['property_set_id'])
        self.assertIsNotNone(process.property_set_id)
        self.assertEqual(process.input_data['property_set_id'], process.property_set_id)

    def test_process_list_of_samples_and_property_sets(self):
        process = self.process
        self.assertIsNotNone(process.output_samples)
        self.assertEqual(len(process.output_samples), 1)
        self.assertEqual(process.output_samples[0].id, self.sample.id)
        samples = process.output_samples
        list = process.make_list_of_samples_for_measurement(samples)
        self.assertTrue(len(list), 1)
        self.assertIsNotNone(list[0]['property_set_id'])
        self.assertIsNotNone(list[0]['sample'])
        self.assertIsNotNone(list[0]['sample'].id)
        self.assertIsNotNone(list[0]['sample'].name)
        self.assertEqual(list[0]['sample'].id, self.sample.id)
        self.assertEqual(list[0]['sample'].name, self.sample.name)

    def test_process_setup_epma_process_as_expected(self):
        template_table = self.make_template_table()
        template_id = self.template_id_with(template_table, 'EPMA')
        epma_process = self.experiment.create_process_from_template(template_id)
        epma_process.rename("EPMA Process For Test")
        epma_process.add_input_samples_to_process([self.sample])
        epma_process.set_value_of_setup_property('voltage', 10)
        epma_process.set_unit_of_setup_property('voltage', 'kV')
        epma_process.set_value_of_setup_property('beam_current', 10)
        epma_process.set_unit_of_setup_property('beam_current', 'nA')
        epma_process.set_value_of_setup_property('beam_size', 0)
        scan_type = 'line'
        epma_process.set_value_of_setup_property('scan_type', scan_type)
        epma_process.set_value_of_setup_property('step_size', 10)
        epma_process.set_value_of_setup_property('grid_dimensions', '20 x 20')
        epma_process.set_value_of_setup_property('location', 'edge')
        epma_process.update_setup_properties([
            'voltage', 'beam_current', 'beam_size', 'scan_type', 'step_size', 'grid_dimensions', 'location'
        ])

    def test_process_setup_epma_process_errors(self):
        pass

    def make_template_table(self):
        template_list = get_all_templates()
        table = {}
        for template in template_list:
            table[template.id] = template
        return table

    def template_id_with(self, table, match):
        found_id = None
        for key in table:
            if match in key:
                found_id = key
        return found_id
